# Copyright 2013 Answers for AWS LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import boto

from boto import ec2

from distami.exceptions import * 
from distami import utils 

__all__ = ('Distami', 'Logging')
log = logging.getLogger(__name__)


class Distami(object):
    def __init__(self, ami_id, ami_region):
        self._ami_id = ami_id
        self._ami_region = ami_region
        
        log.info("Looking for AMI %s in region %s", self._ami_id, self._ami_region)
        try:
            self._conn = ec2.connect_to_region(self._ami_region)
        except boto.exception.NoAuthHandlerFound:
            log.error('Could not connect to region %s' % self._ami_region)
            log.critical('No AWS credentials found. To configure Boto, please read: http://boto.readthedocs.org/en/latest/boto_config_tut.html')
            raise DistamiException('No AWS credentials found.')            
        self._image = utils.wait_for_ami_to_be_available(self._conn, self._ami_id)
        log.debug('AMI details: %s', vars(self._image))
        
        # Get current launch permissions
        self._launch_perms = self._image.get_launch_permissions()
        log.debug("Current launch permissions: %s", self._launch_perms)
        
        # Figure out the underlying snapshot
        ami = utils.get_ami(self._conn, self._ami_id)
        bdm = self._image.block_device_mapping[ami.root_device_name]
        log.debug('Block device mapping for %s: %s', ami.root_device_name, vars(bdm))
        self._snapshot_id = bdm.snapshot_id

        log.info("Found AMI %s with snapshot %s", self._ami_id, self._snapshot_id)

    
    def make_ami_public(self):
        ''' Adds the 'all' group permission to the AMI, making it publicly accessible '''
        
        log.info('Making AMI %s public', self._ami_id)

        if 'groups' in self._launch_perms and any("all" in groups for groups in self._launch_perms['groups']):
            log.debug('AMI already public, nothing to do')
            return True
        
        res = self._image.set_launch_permissions(group_names='all')
        self._launch_perms = self._image.get_launch_permissions()
        return res
    
    
    def make_ami_non_public(self):
        ''' Removes the 'all' group permission from the AMI '''

        log.info('Making AMI %s non-public', self._ami_id)

        if 'groups' in self._launch_perms and any("all" in groups for groups in self._launch_perms['groups']):
            res = self._image.remove_launch_permissions(group_names='all')
            self._launch_perms = self._image.get_launch_permissions()
            return res
        
        log.debug('AMI is already not public')
        return True
    
    def share_ami_with_accounts(self, account_ids):
        ''' Shares an AMI with the supplied list of AWS Account IDs '''
        
        log.info('Sharing AMI %s with AWS Accounts %s', self._ami_id, account_ids)

        res = self._image.set_launch_permissions(user_ids=account_ids)
        self._launch_perms = self._image.get_launch_permissions()
        return res
       
    def make_snapshot_public(self):
        ''' Makes a snapshot public '''
        
        snapshot = utils.get_snapshot(self._conn, self._snapshot_id)
        log.debug('Snapshot details: %s', vars(snapshot))

        log.info('Making snapshot %s public', self._snapshot_id)
        return snapshot.share(groups=['all'])
    
    
    def make_snapshot_non_public(self):
        ''' Removes the 'all' group permission from the snapshot '''
        
        snapshot = utils.get_snapshot(self._conn, self._snapshot_id)
        log.debug('Snapshot details: %s', vars(snapshot))
        
        log.info('Making snapshot %s non-public', self._snapshot_id)
        return snapshot.unshare(groups=['all'])


    def share_snapshot_with_accounts(self, account_ids):
        ''' Shares a snapshot with the supplied list of AWS Account IDs '''
        
        snapshot = utils.get_snapshot(self._conn, self._snapshot_id)
        log.debug('Snapshot details: %s', vars(snapshot))

        log.info('Sharing snapshot %s with AWS Accounts %s', self._snapshot_id, account_ids)
        return snapshot.share(user_ids=account_ids)
    
    
    def copy_to_region(self, region):
        ''' Copies this AMI to another region '''
        
        dest_conn = ec2.connect_to_region(region)
        log.info('Copying AMI to %s', region)
        cp_ami = dest_conn.copy_image(self._ami_region, self._ami_id, self._image.name, self._image.description)
        copied_ami_id = cp_ami.image_id
        
        # Wait for AMI to finish copying before returning
        copied_image = utils.wait_for_ami_to_be_available(dest_conn, copied_ami_id)
        
        # Copy AMI tags to new AMI
        log.info('Copying tags to %s in %s', copied_ami_id, region)
        if self._image.tags:
            dest_conn.create_tags(copied_ami_id, self._image.tags)
        else:
            log.info('AMI tags empty, nothing to copy')
        
        # Also copy snapshot tags to new snapshot
        ami = utils.get_ami(self._conn, self._ami_id)
        copied_snapshot_id = copied_image.block_device_mapping[ami.root_device_name].snapshot_id
        snapshot = utils.get_snapshot(self._conn, self._snapshot_id)
        log.info('Copying tags to %s in %s', copied_snapshot_id, region)

        if snapshot.tags:
            dest_conn.create_tags(copied_snapshot_id, snapshot.tags)
        else:
            log.info('Snapshot tags empty, nothing to copy')

        log.info('Copy to %s complete', region)
        return copied_ami_id



class Logging(object):
    # Logging formats
    _log_simple_format = '%(asctime)s [%(levelname)s] %(message)s'
    _log_detailed_format = '%(asctime)s [%(levelname)s] [%(name)s(%(lineno)s):%(funcName)s] %(message)s'
    
    def configure(self, verbosity = None):
        ''' Configure the logging format and verbosity '''
        
        # Configure our logging output
        if verbosity >= 2:
            logging.basicConfig(level=logging.DEBUG, format=self._log_detailed_format, datefmt='%F %T')
        elif verbosity >= 1:
            logging.basicConfig(level=logging.INFO, format=self._log_detailed_format, datefmt='%F %T')
        else:
            logging.basicConfig(level=logging.INFO, format=self._log_simple_format, datefmt='%F %T')
    
        # Configure Boto's logging output
        if verbosity >= 4:
            logging.getLogger('boto').setLevel(logging.DEBUG)
        elif verbosity >= 3:
            logging.getLogger('boto').setLevel(logging.INFO)
        else:
            logging.getLogger('boto').setLevel(logging.CRITICAL)    
    



# Copyright 2013 Answers for AWS LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import sys

from distami.core import Distami, Logging
from distami import __version__, utils
from distami.exceptions import DistamiException

from boto.utils import get_instance_metadata

from multiprocessing import Pool


__all__ = ('run', )
log = logging.getLogger(__name__)


def _fail(message="Unknown failure", code=1):
    log.error(message)
    sys.exit(code)


def copy(param_array):
    ''' Copies distami to the given region '''
    
    distami = param_array[0]
    to_region = param_array[1]
    args = param_array[2]
    copied_ami_id = distami.copy_to_region(to_region)
    ami_cp = Distami(copied_ami_id, to_region)

    if args.non_public:
        distami.make_ami_non_public()
        distami.make_snapshot_non_public()
    else:
        ami_cp.make_ami_public()
        ami_cp.make_snapshot_public()

    if args.accounts:
        distami.share_ami_with_accounts(args.accounts)
    

def run():
    parser = argparse.ArgumentParser(description='Distributes an AMI by copying it to one, many, or all AWS regions, and by optionally making the AMIs and Snapshots public or shared with specific AWS Accounts.')
    parser.add_argument('ami_id', metavar='AMI_ID', 
                        help='the source AMI ID to distribute. E.g. ami-1234abcd')
    parser.add_argument('--region', metavar='REGION', 
                        help='the region the AMI is in (default is current region of EC2 instance this is running on). E.g. us-east-1')
    parser.add_argument('--to', metavar='REGIONS', 
                        help='comma-separated list of regions to copy the AMI to. The default is all regions. Specify "none" to prevent copying to other regions. E.g. us-east-1,us-west-1,us-west-2')
    parser.add_argument('--non-public', action='store_true', default=False, 
                        help='Copies the AMIs to other regions, but does not make the AMIs or snapshots public. Bad karma, but good for AMIs that need to be private/internal only')
    parser.add_argument('--accounts', metavar='AWS_ACCOUNT_IDs', 
                        help='comma-separated list of AWS Account IDs to share an AMI with. Assumes --non-public. Specify --to=none to share without copying.')
    parser.add_argument('-p', '--parallel', action='store_true', default=False, 
                        help='Perform each copy to another region in parallel. The default is in serial which can take a long time')
    parser.add_argument('-v', '--verbose', action='count', 
                        help='enable verbose output (-vvv for more)')
    parser.add_argument('--version', action='version', version='%(prog)s ' + __version__,
                        help='display version number and exit')
    args = parser.parse_args()
    
    # Argument manipulation
    if args.accounts:
        args.non_public = True
    
    Logging().configure(args.verbose)

    log.debug("CLI parse args: %s", args)

    if args.region:
        ami_region = args.region
    else:
        # If no region was specified, assume this is running on an EC2 instance
        # and work out what region it is in
        log.debug("Figure out which region I am running in...")
        instance_metadata = get_instance_metadata(timeout=5)
        log.debug('Instance meta-data: %s', instance_metadata)
        if not instance_metadata:
            _fail('Could not determine region. This script is either not running on an EC2 instance, or the meta-data service is down')
        
        ami_region = instance_metadata['placement']['availability-zone'][:-1]
        log.debug("Running in region: %s", ami_region)

    try:
        distami = Distami(args.ami_id, ami_region)
        if not args.non_public:
            distami.make_ami_public()
            distami.make_snapshot_public()
        if args.accounts:
            account_ids = args.accounts.split(',')
            distami.share_ami_with_accounts(account_ids)
            distami.share_snapshot_with_accounts(account_ids)
        
        if args.to and args.to == 'none':
            to_regions = []
        elif args.to and args.to != 'all':
            # TODO It is probably worth sanity checking this for typos
            to_regions = args.to.split(',')
        else:
            to_regions = utils.get_regions_to_copy_to(ami_region)
        
        if args.parallel:
            # Get input to copy function 
            f = lambda x,y,z: [x, y, z]
            param_array = map(f, [distami] * len(to_regions), to_regions, [args] * len(to_regions))
            log.debug(param_array)

            # Copy to regions in parallel
            log.info('Copying in parallel. Hold on to your hat...')
            pool = Pool(processes=8)
            pool.map(copy, param_array)
            pool.close()
            pool.join()
        else:
            # Copy to regions one at a time
            for region in to_regions:
                copy([distami, region, args])
        
    except DistamiException as e:
        _fail(e.message)
    
    log.info('AMI successfully distributed!')
    sys.exit(0)
    



# Copyright 2013 Answers for AWS LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import time
import boto
from boto import ec2

from distami.exceptions import * 

log = logging.getLogger(__name__)


def get_ami(conn, ami_id):
    ''' Gets a single AMI as a boto.ec2.image.Image object '''
    
    attempts = 0
    max_attempts = 5
    while (attempts < max_attempts):
        try:
            attempts += 1
            images = conn.get_all_images(ami_id)
        except boto.exception.EC2ResponseError:
            msg = "Could not find AMI '%s' in region '%s'" % (ami_id, conn.region.name)
            if attempts < max_attempts:
                # The API call to initiate an AMI copy is not blocking, so the
                # copied AMI may not be available right away
                log.debug(msg + ' so waiting 5 seconds and retrying')
                time.sleep(5)
            else:
                raise DistamiException(msg)
    
    log.debug("Found AMIs: %s", images)
    if len(images) != 1:
        msg = "Somehow more than 1 AMI was detected - this is a weird error"
        raise DistamiException(msg)
        
    return images[0]


def get_snapshot(conn, snapshot_id):
    ''' Gets a single snapshot as a boto.ec2.snapshot.Snapshot object '''
    
    try:
        snapshots = conn.get_all_snapshots(snapshot_id)
    except boto.exception.EC2ResponseError:
        msg = "Could not snapshot '%s' in region '%s'" % (snapshot_id, conn.region.name)
        raise DistamiException(msg)
    
    log.debug("Found snapshots: %s", snapshots)
    if len(snapshots) != 1:
        msg = "Somehow more than 1 snapshot was detected - this is a weird error"
        raise DistamiException(msg)
        
    return snapshots[0]


def get_regions_to_copy_to(source_region):
    ''' Gets the list of regions to copy an AMI to '''
    
    regions = []
    for region in ec2.regions():
        if region.name == source_region:
            continue
        # Filter out GovCloud
        if region.name == 'us-gov-west-1':
            continue
        # Filter out China
        if region.name == 'cn-north-1':
            continue
        regions.append(region.name)
        
    return regions


def wait_for_ami_to_be_available(conn, ami_id):
    ''' Blocking wait until the AMI is available '''
    
    ami = get_ami(conn, ami_id)
    log.debug('AMI details: %s', vars(ami))
    
    while ami.state != 'available':
        log.info("%s in %s not available, waiting...", ami_id, conn.region.name)
        time.sleep(30)
        ami = get_ami(conn, ami_id)
        
        if ami.state == 'failed':
            msg = "AMI '%s' is in a failed state and will never be available" % ami_id
            raise DistamiException(msg)
    
    return ami


#!/usr/bin/env python
# Copyright 2013 Answers for AWS LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
setuptools install script for DistAMI
"""

import sys
major, minor = sys.version_info[0:2]
if major != 2 or minor < 7:
    print 'DistAMI requires Python 2.7.x'
    sys.exit(1)

from setuptools import setup, find_packages

import distami

with open('requirements.txt') as fh:
    requires = [requirement.strip() for requirement in fh]

entry_points = {
    'console_scripts': [
        'distami = distami.cli:run',
    ]
}

exclude_packages = [
    'tests',
    'tests.*',
]

setup(
    name='distami',
    version=distami.__version__,
    description='Distribute Amazon Machine Images (AMIs) to the public',
    long_description=open('README.rst').read(),
    author=distami.__author__,
    author_email='info@answersforaws.com',
    url='https://github.com/Answers4AWS/distami',
    packages=find_packages(exclude=exclude_packages),
    package_dir={'distami': 'distami'},
    include_package_data=True,
    zip_safe=False,
    install_requires=requires,
    entry_points=entry_points,
    license=open("LICENSE.txt").read(),
    classifiers=(
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Topic :: System :: Installation/Setup',
        'Topic :: Utilities',
    )
)
